#include <stdio.h>
#include <math.h>

#include <gl/glew.h>
#include <gl/freeglut.h>

#include <cuda_runtime.h>

#define INIT_X_POS 128
#define INIT_Y_POS 128
#define INIT_WIDTH 512
#define INIT_HEIGHT 512

unsigned int window_width;
unsigned int window_height;

double init_left = -0.25;
double init_right = 1.25;
double init_bottom = -0.25;
double init_top = 1.25;
double left, right, bottom, top;

// qƂ̈ʒuD
#define NUM_POINTS (1024 * 1024)
float h_point[NUM_POINTS][2];
float (*d_point)[2];

// _obt@IuWFNgD
GLuint vbo;

// ԂƎԍ݁D
float anim_time = 0.0f;
float anim_dt = 0.01f;

extern void launchGPUKernel(unsigned int num_particles, float (*pos)[2], float time, float dt);
extern void launchCPUKernel(unsigned int num_particles, float (*pos)[2], float time, float dt);

// qʒuɔzuD
void setInitialPosition(void)
{
	unsigned int i;
	srand(12131);
	for (i = 0; i < NUM_POINTS; i++) {
		h_point[i][0] = (float)rand() / RAND_MAX * 0.5f + 0.25f;
		h_point[i][1] = (float)rand() / RAND_MAX * 0.5f;
	}

	// GPUɃf[^̏ʒu]D
	cudaMalloc((void**)&d_point, NUM_POINTS * 2 * sizeof(float));
	cudaMemcpy(d_point, h_point,  NUM_POINTS * 2 * sizeof(float), cudaMemcpyHostToDevice);
}

// _obt@IuWFNg̐D
void createVBO(GLuint *vbo, unsigned int size)
// GLuint *bo; _obt@IuWFNgD
// unsigned int size; obt@̃TCYD
{
    glGenBuffers(1, vbo);

    // _f[^̒_obt@IuWFNgւ̐ݒD
    glBindBuffer(GL_ARRAY_BUFFER, *vbo);
	glBufferData(GL_ARRAY_BUFFER, size, h_point, GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

// _obt@IuWFNg̍폜D
void deleteVBO(GLuint *vbo)
// GLuint *bo; _obt@IuWFNgD
{
    glDeleteBuffers(1, vbo);
    *vbo = 0;
}

// CPUɂ́D
void runCPUKernel(void)
{

	// CPǓĂяoD
    launchCPUKernel(NUM_POINTS, h_point, anim_time, anim_dt);
	anim_time += anim_dt;
}

// GPUɂ́D
void runGPUKernel(void)
{

	// GPǓĂяoD
    launchGPUKernel(NUM_POINTS, d_point, anim_time, anim_dt);
	cudaMemcpy(h_point, d_point, NUM_POINTS * 2 * sizeof(float), cudaMemcpyDeviceToHost);

	// _f[^̒_obt@IuWFNgւ̍ĐݒD
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, NUM_POINTS * 2 * sizeof(float), h_point, GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
	anim_time += anim_dt;
}

// \D
void display(void)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	glOrtho(left, right, bottom, top, -100.0, 100.0);
	glViewport(0, 0, window_width, window_height);

    // qʒu̍XV
	//runCPUKernel(); // CPUD
	runGPUKernel(); // GPUD

    // _obt@IuWFNgɊÂ_̕`D
	glClear(GL_COLOR_BUFFER_BIT);
	glColor3f(1.0f, 0.0f, 0.0f);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FLOAT, 0, 0);
	glDrawArrays(GL_POINTS, 0, NUM_POINTS);
    glDisableClientState(GL_VERTEX_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	// 摜̍XVD
    glutSwapBuffers();
    glutPostRedisplay();
}

// TCYD
void resize(int width, int height)
{
	double dx, dy, d_aspect, w_aspect, d;

	// EBhETCY̎擾D
	window_width = width;
	window_height = height;

	// \͈͂̃AXyNgƃEBhẼAXyNg̔rD
	dx = init_right - init_left;
	dy = init_top - init_bottom;
	d_aspect = dy / dx;
	w_aspect = (double)window_height / (double)window_width;
	
	// EBhE\͈͂cD\͈͂cɍLD
	if (w_aspect > d_aspect) {
		d = (dy * (w_aspect / d_aspect - 1.0)) * 0.5;
		left = init_left;
		right = init_right;
		bottom = init_bottom - d;
		top = init_top + d;

		// EBhE\͈͂D\͈͂ɍLD
	} else {
		d = (dx * (d_aspect / w_aspect - 1.0)) * 0.5;
		left = init_left - d;
		right = init_right + d;
		bottom = init_bottom;
		top = init_top;
	}
}

// ㏈D
void cleanUp(void)
{
	cudaFree(d_point);

	// _obt@IuWFNg̏D
    deleteVBO(&vbo);
	cudaDeviceReset();
}

// L[D
void keyboard(unsigned char key, int x, int y)
{
	switch (key) {
		case 'q':
		case 'Q':
		case '\033':
			exit(0);
    }
}

// OpenGL֌W̏ݒD
bool initGL(void)
{
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	// glew̏D
    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0")) {
		fprintf(stderr, "ERROR: Support for necessary OpenGL extensions missing.");
		return false;
    }
  
    // _obt@IuWFNg̐D
    createVBO(&vbo, NUM_POINTS * 2 * sizeof(float));
 	return true;
}

int main(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowPosition(INIT_X_POS, INIT_Y_POS);
	glutInitWindowSize(INIT_WIDTH, INIT_HEIGHT);
    glutCreateWindow("2D Particle Simulation");
    glutDisplayFunc(display);
	glutReshapeFunc(resize);
	glutKeyboardFunc(keyboard);
    atexit(cleanUp);

    // qʒuɔzuD
    setInitialPosition();
 
	// OpenGL̐ݒD
    if (!initGL())
        return 1;
 
    // Aj[V`̃[vD
    glutMainLoop();
	return 0;
}
