#include <math.h>

#define PI 3.141592653589793

// CPUD
#define h_U(x, y, t) (- 2.0f * (float)cos(PI * (t) / 8.0f) * (float)sin(PI * (x)) * (float)sin(PI * (x)) * (float)cos(PI * (y)) * (float)sin(PI * (y)))
#define h_V(x, y, t) (2.0f * (float)cos(PI * (t) / 8.0f) * (float)cos(PI * (x)) * (float)sin(PI * (x)) * (float)sin(PI * (y)) * (float)sin(PI * (y)))

// CPUpQENb^@D
void h_RungeKutta(unsigned int index, float (*pos)[2], float time, float dt)
// unsigned int index; q̃CfbNXD
// float (*pos)[2]; qʒuD
// float time; D
// float dt; ԍ݁D
{
 	float xn, yn, p1, q1, p2, q2, p3, q3, p4, q4;
	float x, y, t;
    xn = pos[index][0];
    yn = pos[index][1];

    // 1iځD
	p1 = h_U(xn, yn, time);
	q1 = h_V(xn, yn, time);

	// 2iځD
	x = xn + 0.5f * p1 * dt;
	y = yn + 0.5f * q1 * dt;
	t = time + 0.5f * dt;
	p2 = h_U(x, y, t);
	q2 = h_V(x, y, t);

	// 3iځD
	x = xn + 0.5f * p2 * dt;
	y = yn + 0.5f * q2 * dt;
	t = time + 0.5f * dt;
	p3 = h_U(x, y, t);
	q3 = h_V(x, y, t);

	// 4iځD
	x = xn + p3 * dt;
	y = yn + q3 * dt;
	t = time + dt;
	p4 = h_U(x, y, t);
	q4 = h_V(x, y, t);
   
    // qʒu̍XVD
    pos[index][0] = xn + (p1 + 2 * p2 + 2 * p3 + p4) / 6.0f * dt;
    pos[index][1] = yn + (q1 + 2 * q2 + 2 * q3 + q4) / 6.0f * dt;
}

// CPŰND
void launchCPUKernel(unsigned int num_particles, float (*pos)[2], float time, float dt)
// unsigned int num_particles; q̑D
// float (*pos)[2]; qʒuD
// float time; D
// float dt; ԍ݁D
{
    unsigned int i;
    for (i = 0; i < num_particles; i++)
        h_RungeKutta(i, pos, time, dt);
}

// GPUD
#define d_U(x, y, t) (- 2.0f * __cosf(PI * (t) / 8.0f) * __sinf(PI * (x)) * __sinf(PI * (x)) * __cosf(PI * (y)) * __sinf(PI * (y)))
#define d_V(x, y, t) (2.0f * __cosf(PI * (t) / 8.0f) * __cosf(PI * (x)) * __sinf(PI * (x)) * __sinf(PI * (y)) * __sinf(PI * (y)))

// GPUpQENb^@
__global__ void d_RungeKutta(unsigned int num_particles, float (*pos)[2], float time, float dt)
// unsigned int num_particles; q̑D
// float (*pos)[2]; qʒuD
// float time; D
// float dt; ԍ݁D
{
    unsigned int index;
	float xn, yn, p1, q1, p2, q2, p3, q3, p4, q4;
	float x, y, t;

    // Ώۂ̗q̌D
    index = blockDim.x * blockIdx.x + threadIdx.x;
    if (index >= num_particles)
		return;
	xn = pos[index][0];
    yn = pos[index][1];

    // 1iځD
	p1 = d_U(xn, yn, time);
	q1 = d_V(xn, yn, time);

	// 2iځD
	x = xn + 0.5f * p1 * dt;
	y = yn + 0.5f * q1 * dt;
	t = time + 0.5f * dt;
	p2 = d_U(x, y, t);
	q2 = d_V(x, y, t);

	// 3iځD
	x = xn + 0.5f * p2 * dt;
	y = yn + 0.5f * q2 * dt;
	t = time + 0.5f * dt;
	p3 = d_U(x, y, t);
	q3 = d_V(x, y, t);

	// 4iځD
	x = xn + p3 * dt;
	y = yn + q3 * dt;
	t = time + dt;
	p4 = d_U(x, y, t);
	q4 = d_V(x, y, t);
   
    // qʒu̍XVD
    pos[index][0] = xn + (p1 + 2 * p2 + 2 * p3 + p4) / 6.0f * dt;
    pos[index][1] = yn + (q1 + 2 * q2 + 2 * q3 + q4) / 6.0f * dt;
}

// GPŰND
void launchGPUKernel(unsigned int num_particles, float (*pos)[2], float time, float dt)
// unsigned int num_particles; q̑D
// float (*pos)[2]; qʒuD
// float time; D
// float dt; ԍ݁D
{
    dim3 grid(num_particles / 512 + 1, 1);
    dim3 block(512, 1, 1);
    d_RungeKutta <<< grid, block >>> (num_particles, pos, time, dt);
}