#include <stdio.h>

#include <gl/glew.h>
#include <gl/freeglut.h>

#include <cuda_runtime.h>

#define INIT_X_POS 128
#define INIT_Y_POS 128
#define INIT_WIDTH 512
#define INIT_HEIGHT 512

unsigned int window_width = INIT_WIDTH;
unsigned int window_height = INIT_HEIGHT;

#define GRID_WIDTH 1024
#define GRID_HEIGHT 1024

// Zxli[2iqD
float *h_F, *h_Fn;
float *d_F, *d_Fn;
float dx, dy;

// iqԂ͈̔́D
float min_x = -0.5f;
float max_x = 0.5f;
float min_y = -0.5f;
float max_y = 0.5f;

// iqi摜j̉𑜓xD
unsigned int grid_width = GRID_WIDTH;
unsigned int grid_height = GRID_HEIGHT;

// eNX`D
GLuint tex;
int *h_g_data;
int *d_g_data;

// sNZobt@IuWFNgD
GLuint pbo;
struct cudaGraphicsResource *pbo_res;

// JԂ񐔂̃JE^D
unsigned int incrt = 1;

// ͎ԂƎԍ݁D
float anim_time = 0.0f;
float anim_dt;

// ͂ŗpp[^D
float kappa = 0.1f;
float max_density;

extern void launchCPUKernel(float *F, float *Fn, unsigned int width, unsigned int height, float dx, float dy, float kappa, float dt,
	float max_density, int *g_data);
extern void launchGPUKernel(float *F, float *Fn, unsigned int width, unsigned int height, float dx, float dy, float kappa, float dt,
    float max_density, bool direction, int *g_data);
extern void bindTextures(float *F, float *Fn);
extern void unbindTextures(void);

// Zx̐ݒD
void setInitialDensity(unsigned int width, unsigned int height, float *max_density)
// unsigned int width, height; iqXỶ𑜓xD
// float *max_density; Zx̍őlD
{
    unsigned int i, j, index;
    float x, y;
	dx = (max_x - min_x) / width;
	dy = (max_y - min_y) / height;

	// `̔Zx^D
	*max_density = 1.0f;
 	h_F = (float *)malloc(width * height * sizeof(float));
	h_Fn = (float *)malloc(width * height * sizeof(float));
    for (i = 0; i < height; i++) {
        for (j = 0; j < width; j++) {
            index = width * i + j;
            x = dx * (j + 0.5f) + min_x;
            y = dy * (i + 0.5f) + min_y;
			if ((x > (0.75f * min_x + 0.25f * max_x)) && (x < (0.25f * min_x + 0.75f * max_x)) 
				&& (y > (0.75f * min_y + 0.25f * max_y)) && (y < (0.25f * min_y + 0.75f * max_y)))
				h_F[index] = *max_density;
			else
				h_F[index] = 0.0f;
        }
    }

	// GPUp̃̊mہD
    cudaMalloc((void**)&d_F, width * height * sizeof(float));
	cudaMalloc((void**)&d_Fn, width * height * sizeof(float));

    // h_F̃f[^foCX̔zd_F֓]D
    cudaMemcpy(d_F, h_F, width * height * sizeof(float), cudaMemcpyHostToDevice);

	// CUDAeNX`̃oChD
	bindTextures(d_F, d_Fn);
}

// eNX`̐D
void createTexture(GLuint *tex, unsigned int width, unsigned int height)
// GLuint *tex; eNX`D
// unsinged int width, height; eNX`Ɋi[C[Wf[^̉𑜓xD
{
    glGenTextures(1, tex);

    // {Iȃp[^̐ݒD
    glBindTexture(GL_TEXTURE_2D, *tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    // i[C[Wf[^̒`D
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
}

// eNX`̍폜D
void deleteTexture(GLuint* tex)
// GLuint *tex; eNX`D
{
    glDeleteTextures(1, tex);
    *tex = 0;
}

bool direction_p = true;

// CPUɂ́D
void runCPUKernel(void) 
{

	// CPǓĂяoD
	if (direction_p)
		launchCPUKernel(h_F, h_Fn, grid_width, grid_height, dx, dy, kappa, anim_dt, max_density, h_g_data);
	else
		launchCPUKernel(h_Fn, h_F, grid_width, grid_height, dx, dy, kappa, anim_dt, max_density, h_g_data);
	direction_p = !direction_p;
	anim_time += anim_dt;
	
	// 摜f[^eNX`֓]D	
	glBindTexture(GL_TEXTURE_2D, tex);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, grid_width, grid_height, GL_RGBA, GL_UNSIGNED_BYTE, h_g_data);
}

// GPUɂ́D
void runGPUKernel(void) 
{

	// GPǓĂяoD
	if (direction_p)
		launchGPUKernel(d_F, d_Fn, grid_width, grid_height, dx, dy, kappa, anim_dt, max_density, direction_p, d_g_data);
	else
		launchGPUKernel(d_Fn, d_F, grid_width, grid_height, dx, dy, kappa, anim_dt, max_density, direction_p, d_g_data);
	direction_p = !direction_p;
	anim_time += anim_dt;
		
	// 摜f[^eNX`֓]D
	cudaMemcpy(h_g_data, d_g_data, grid_width * grid_height * sizeof(int), cudaMemcpyDeviceToHost);
	glBindTexture(GL_TEXTURE_2D, tex);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, grid_width, grid_height, GL_RGBA, GL_UNSIGNED_BYTE, h_g_data);
}

// \D
void display(void)
{

	// gUۂ̉͂Ɖ摜̐D
	//runCPUKernel(); // CPUD
	runGPUKernel(); // GPUD

	// es̐ݒD
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(- 1.0, 1.0, - 1.0, 1.0, -1.0, 1.0);
    glViewport(0, 0, window_width, window_height);

    // eNX`\tꂽlp`̕`D
	glClear(GL_COLOR_BUFFER_BIT);
	glBindTexture(GL_TEXTURE_2D, tex);
 	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f); glVertex2f(- 1.0, - 1.0); 
    glTexCoord2f(1.0f, 0.0f); glVertex2f(1.0, - 1.0); 
    glTexCoord2f(1.0f, 1.0f); glVertex2f(1.0, 1.0); 
    glTexCoord2f(0.0f, 1.0f); glVertex2f(- 1.0, 1.0);
    glEnd();	
	
	// 摜̍XVD
	glutSwapBuffers();
	glutPostRedisplay();

    // ̐i\D
	incrt++;
    if (incrt % 100 == 0) 
		printf("time(%4d): %7.5f\n", incrt, anim_time);
}

// EBhẼTCYD
void resize(int width, int height)
{
    window_width = width;
    window_height = height;
}

// ㏈D
void cleanUp(void)
{
	free(h_F);
	free(h_Fn);
	cudaFree(d_F);
	cudaFree(d_Fn);

	// CUDAeNX`̃AoChD
	unbindTextures();

	// eNX`̏D
	free(h_g_data);
	cudaFree(d_g_data);
	deleteTexture(&tex);
	cudaDeviceReset();
}

// L[D
void keyboard(unsigned char key, int /*x*/, int /*y*/)
{
	switch (key) {
		case 'q':
		case 'Q':
		case '\033':
			exit(0);
    }
}

// OpenGL֌W̏ݒD
bool initGL(void)
{
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glEnable(GL_TEXTURE_2D);

	// glew̏D
    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0")) {
		fprintf(stderr, "ERROR: Support for necessary OpenGL extensions missing.");
        return false;
    }

	// eNX`̐D
	h_g_data = (int *)malloc(grid_width * grid_height * sizeof(int));
	cudaMalloc((void**)&d_g_data, grid_width * grid_height * sizeof(int));
    createTexture(&tex, grid_width, grid_height);
    return true;
}

int main(int argc, char** argv)
{
 
	// Zx̐ݒƍőZx̎擾D
    setInitialDensity(grid_width, grid_height, &max_density);
	printf("Max: %f\n", max_density);

    // ԍ݂̐ݒD
	anim_dt = 0.2f * min(dx * dx, dy * dy) / kappa;

    // GLUT֌W̐ݒD
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowPosition(INIT_X_POS, INIT_Y_POS);
	glutInitWindowSize(INIT_WIDTH, INIT_HEIGHT);
    glutCreateWindow("2D Diffusion Simulation");
    glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
    glutReshapeFunc(resize);
	atexit(cleanUp);

	// OpenGL̐ݒD
	if (!initGL())
        return 1;

    // Aj[V`̃[vD
    glutMainLoop();
    return 0;
}
