#include <cuda_runtime.h>

#include <math.h>
#include <helper_math.h>

#include "Metaball.h"

// thrustp̃wb_[t@CD
#include <thrust/scan.h>
#include <thrust/device_ptr.h>

// iq\̃p[^ێ\́D
__constant__ struct sim_params d_params;

// p[^ݒD
void setParameters(struct sim_params *h_params)
{
    cudaMemcpyToSymbol(d_params, h_params, sizeof(struct sim_params), 0, cudaMemcpyHostToDevice);
}

// 1ObhTCY̊ȈՌvZD
void computeGridSize(uint n, uint block_size, uint *grid_size, uint *new_block_size)
// unit n; Ώۂ̑D
// unit block_size; 1ubÑXbhD
// uint &grid_size; ObhTCYD
// uint &new_block_size; ␳ꂽ1ubÑXbhD
{
    *new_block_size = min(block_size, n);
    *grid_size = ((n % *new_block_size == 0) ? (n / *new_block_size) : (n / *new_block_size + 1));
}

// W3iqɂObhʒuD
__device__ int3 d_pointToGridPos(float4 point)
// float4 point; WlD
{
    int3 grid_pos;
    grid_pos.x = (int)floor((point.x - d_params.world_origin.x) / d_params.voxel_size.x);
    grid_pos.y = (int)floor((point.y - d_params.world_origin.y) / d_params.voxel_size.y);
    grid_pos.z = (int)floor((point.z - d_params.world_origin.z) / d_params.voxel_size.z);
    return grid_pos;
}

// ObhʒuɂWD
__device__ float4 d_gridPosToPoint(int3 grid_pos)
// int3 grid_pos; ObhʒuD
{
    float4 p;
    p.x = d_params.world_origin.x + grid_pos.x * d_params.voxel_size.x;
    p.y = d_params.world_origin.y + grid_pos.y * d_params.voxel_size.y;
    p.z = d_params.world_origin.z + grid_pos.z * d_params.voxel_size.z;
    p.w = 1.0f;
    return p;
}    

// 3IȃObhʒu1IȃZAhXւ̕ϊD
__device__ uint d_gridPosToAddress(int3 grid_pos)
// int3 grid_pos; ObhʒuD
{

    // ʒuiq͈͓̔Ɏ܂悤ɕ␳D
    if (grid_pos.x < 0) grid_pos.x = 0;
    if (grid_pos.x > (d_params.grid_size.x - 1)) grid_pos.x = (d_params.grid_size.x - 1);
    if (grid_pos.y < 0) grid_pos.y = 0;
    if (grid_pos.y > (d_params.grid_size.y - 1)) grid_pos.y = (d_params.grid_size.y - 1);
    if (grid_pos.z < 0) grid_pos.z = 0;
    if (grid_pos.z > (d_params.grid_size.z - 1)) grid_pos.z = (d_params.grid_size.z - 1);

    // 1IȃZAhX̌vZD
    return __umul24(__umul24(d_params.grid_size.x, d_params.grid_size.y), grid_pos.z) + __umul24(d_params.grid_size.x, grid_pos.y) + grid_pos.x;
}

// 3IȃObhʒuiq͈͓̔ǂ̔D
__device__ bool d_validGridPosP(int3 grid_pos)
// int3 grid_pos; ObhʒuD
{
	return((grid_pos.x >= 0) && (grid_pos.x < d_params.grid_size.x) 
		&& (grid_pos.y >= 0) && (grid_pos.y < d_params.grid_size.y)
		&& (grid_pos.z >= 0) && (grid_pos.z < d_params.grid_size.z));
}

// ZAhXObhʒuւ̕ϊD
__device__ int3 d_addressToGridPos(uint address)
// uint address; ZAhXD
{
	uint tmp;
    int3 grid_pos;
    grid_pos.z = address / (d_params.grid_size.x * d_params.grid_size.y);
    tmp = address % (d_params.grid_size.x * d_params.grid_size.y);
    grid_pos.y = tmp / d_params.grid_size.x;
    grid_pos.x = tmp % d_params.grid_size.x; 
    return grid_pos;
}

// Wւ̃ZAhX̊蓖āD
__global__ void d_calcAddress(uint num_points, float4 *point, uint *grid_point_address, uint *grid_point_index)
// uint num_points; _̑D
// float4 *point; _QD
// unit *grid_point_address; e_3iqɂZAhXD
// unit *grid_point_index; e_̃CfbNXD
{
    uint index, address;

    // Ώۂ̓_̑ID
	index = __umul24(blockDim.x, blockIdx.x) + threadIdx.x;
    if (index >= num_points) 
        return;

    // ZAhX̎擾D
	address = d_gridPosToAddress(d_pointToGridPos(point[index]));

    // ꂽAhX_̃CfbNXƂƂɊi[D
    grid_point_address[index] = address;
    grid_point_index[index] = index;
}

// Ăяo֐D
void launchCalcAddress(uint num_points, float4 *point, uint *grid_point_address, uint *grid_point_index)
// uint num_points; _̑D
// float4 *point; _QD
// unit *grid_point_address; e_3iqɂZAhXD
// unit *grid_point_index; e_̃CfbNXD
{
    uint grid, block;
    computeGridSize(num_points, 256, &grid, &block);
    d_calcAddress <<< grid, block >>> (num_points, point, grid_point_address, grid_point_index);
}

// ZɊi[闱q̐JEgD
__global__ void d_countPointsInVoxel(uint num_points, uint *grid_point_address, uint *voxel_counter)
// uint num_points; _̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_counter; eZɊi[Ă_̌D
{
	uint index;

    // Ώۂ̓_̑ID
	index = __umul24(blockDim.x, blockIdx.x) + threadIdx.x;
    if (index >= num_points) 
        return;

	// Ag~bN֐pāCeZɊi[闱qJEgD
	atomicAdd(&voxel_counter[grid_point_address[index]], 1);
}

// Ăяo֐D
void launchCountPointsInVoxel(uint num_points, uint num_voxels, uint *grid_point_address, uint *voxel_counter)
// uint num_points; _̑D
// uint num_voxels; Z̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_counter; eZɊi[Ă_̌D
{
	uint grid, block;
	
    // \voxel_counteȓSvf0܂ĂD
	cudaMemset(voxel_counter, 0, num_voxels * sizeof(uint));
    computeGridSize(num_points, 256, &grid, &block);
    d_countPointsInVoxel <<< grid, block >>> (num_points, grid_point_address, voxel_counter);
}

// rIXLD
void thrustScan(unsigned int *input, unsigned int num_elements, unsigned int *output)
{
    thrust::exclusive_scan(thrust::device_ptr<unsigned int>(input), 
		thrust::device_ptr<unsigned int>(input + num_elements),
		thrust::device_ptr<unsigned int>(output));
}

// X̗qZɊi[D
__global__ void d_storePointToVoxel(uint num_points, uint *grid_point_address, uint *voxel_end, uint *voxel)
// uint num_points; _̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_end; eZւ̓_̊i[JnꏊD
// uint *voxel; Z\D
{
	uint index, old_count;

    // Ώۂ̓_̑ID
	index = __umul24(blockDim.x, blockIdx.x) + threadIdx.x;
    if (index >= num_points) 
        return;

	// Ag~bN֐p_̃{NZւ̒ǉD
	old_count = atomicAdd(&voxel_end[grid_point_address[index]], 1);
	voxel[old_count] = index;
}

// Ăяo֐D
void launchStorePointToVoxel(uint num_points, uint *grid_point_address, uint *voxel_end, uint *voxel)
// uint num_points; _̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_end; eZւ̓_̊i[JnꏊD
// uint *voxel; Z\D
{
	uint grid, block;
    computeGridSize(num_points, 256, &grid, &block);
    d_storePointToVoxel <<< grid, block >>> (num_points, grid_point_address, voxel_end, voxel);
}

// iq_ɂăZ̗q̔ZxώZD
__device__ float d_accumulatePointsInCell(float4 grid_point, int3 grid_pos, float4 *point, 
	uint *voxel_start, uint *voxel_end, uint *voxel, float limit_dist)
// float4 grid_point; iq̓_D
// int3 grid_pos; Z̃ObhʒuD̃Z̓_̉eώZD
// float4 *point; _QD
// uint *voxel_start, *voxel_end; eZɊi[Ă_voxelɂ͈́D
// uint *voxel; Z\D
// float limit_dist; ł؂苗D
{
	uint i;
    uint address = d_gridPosToAddress(grid_pos);
	float value = 0.0f, d2;
    float4 rel_pos;

    // Zɓ_܂܂ĂD
    if (d_validGridPosP(grid_pos)) {

        // ̃Z̓_̉eώZD
		for (i = voxel_start[address]; i < voxel_end[address]; i++) {

            // _Ɗiq̓_̑ΈʒuvZD
            rel_pos = point[voxel[i]] - grid_point;
            d2 = rel_pos.x * rel_pos.x + rel_pos.y * rel_pos.y + rel_pos.z * rel_pos.z;
            if (d2 < limit_dist * limit_dist) // ł؂苗łΐώZD
                value += 0.0002f / d2;
        }
    }
    return value;
}

// eiq_ւ̔Zxl̊蓖āD
__global__ void d_setDensityToGrid(uint num_voxels, float4 *point, uint *voxel_start, uint *voxel_end, uint *voxel, float limit_dist, float *d_density)
// uint num_voxels; Z̑D
// float4 *point; _QD
// uint *voxel_start, *voxel_end; eZɊi[Ă_voxelɂ͈́D
// uint *voxel; Z\D
// float limit_dist; ł؂苗D
// float *d_density; eiq̔ZxlD
{
    uint block_index = __mul24(gridDim.x, blockIdx.y) + blockIdx.x; 
    uint address = __mul24(blockDim.x, block_index) + threadIdx.x;
	int x, y, z;
    int3 near_cell, grid_range;
	int3 grid_pos;
	float value;
	float4 grid_point;

    // ̃Xbh̃CfbNXiZAhXj̓Z̑I[o[ĂD
    if (address >= num_voxels) 
        return;

	// `FbNZ͈̔́D1]܂D
	grid_range.x = (int)ceil(limit_dist / d_params.voxel_size.x);
	grid_range.y = (int)ceil(limit_dist / d_params.voxel_size.y);
	grid_range.z = (int)ceil(limit_dist / d_params.voxel_size.z);

    // iq_̃Obhʒu𓾂D
    grid_pos = d_addressToGridPos(address);

    // iq̍W̎擾D
    grid_point = d_gridPosToPoint(grid_pos);

    // ߖT̃Z̓_ɂZxlώZD
    value = 0.0f;
    for (z = - grid_range.z; z < grid_range.z; z++) {
        for (y = - grid_range.y; y < grid_range.y; y++) {
            for (x = - grid_range.x; x < grid_range.x; x++) {
                near_cell.x = grid_pos.x + x;
                near_cell.y = grid_pos.y + y;
                near_cell.z = grid_pos.z + z;
                value += d_accumulatePointsInCell(grid_point, near_cell, point, voxel_start, voxel_end, voxel, limit_dist);
            }
        }
    }

    // ώZꂽZxliqɊi[D
    d_density[address] = value;
}

// Ăяo֐D
void launchSetDensityToGrid(uint num_voxels, float4 *point, uint *voxel_start, uint *voxel_end, uint *voxel, float limit_dist, float *d_density)
// uint num_voxels; Z̑D
// float4 *point; _QD
// uint *voxel_start, *voxel_end; eZɊi[Ă_voxelɂ͈́D
// uint *voxel; Z\D
// float limit_dist; ł؂苗D
// float *d_density; eiq̔ZxlD
{
    int block = 256;
    dim3 grid(num_voxels / block + 1, 1);
    if (grid.x > 65535) {
        grid.y = grid.x / 32768 + 1;
        grid.x = 32768;
    }
    d_setDensityToGrid <<< grid, block >>> (num_voxels, point, voxel_start, voxel_end, voxel, limit_dist, d_density);
}
